@@ -6,19 +6,107 @@ property surface of a system.
 from __future__ import division
 from pycalphad import Model
 from pycalphad.model import DofError
-from pycalphad.core.autograd_utils import build_functions
-from pycalphad.core.tempfilemanager import TempfileManager
+from pycalphad.core.sympydiff_utils import build_functions
 from pycalphad.core.utils import point_sample, generate_dof
 from pycalphad.core.utils import endmember_matrix, unpack_kwarg
 from pycalphad.core.utils import broadcast_to, unpack_condition, unpack_phases
+from pycalphad.core.cache import cacheit
+from pycalphad.core.phase_rec import PhaseRecord, PhaseRecord_from_cython, PhaseRecord_from_compiledmodel
+from pycalphad.core.compiled_model import CompiledModel
 from pycalphad.log import logger
 import pycalphad.variables as v
 from sympy import Symbol
-from xarray import concat, Dataset, DataArray
 import numpy as np
 import itertools
 import collections
-import os
+from xarray import Dataset, Variable
+from xarray.core.dataset import as_dataset
+from xarray.core.combine import _calc_concat_over, _calc_concat_dim_coord, concat_vars
+from collections import OrderedDict
+import pandas as pd
+
+class FallbackModel(object):
+    "Compatibility layer while transitioning to CompiledModel."
+    def __new__(cls, *args, **kwargs):
+        try:
+            ret = CompiledModel(*args, **kwargs)
+        except NotImplementedError:
+            return Model(*args, **kwargs)
+        return ret
+
+def _fast_concat(datasets, dim=None):
+    """
+    Fork of xarray.concat for Datasets.
+    This function assumes everything is already aligned, and skips error checking.
+
+    Parameters
+    ----------
+    datasets
+
+    Returns
+    -------
+
+    """
+    data_vars = 'all'
+    coords = 'different'
+    positions = None
+
+    dim, coord = _calc_concat_dim_coord(dim)
+    datasets = [as_dataset(ds) for ds in datasets]
+
+    concat_over = _calc_concat_over(datasets, dim, data_vars, coords)
+
+    def insert_result_variable(k, v):
+        assert isinstance(v, Variable)
+        if k in datasets[0].coords:
+            result_coord_names.add(k)
+        result_vars[k] = v
+
+    # create the new dataset and add constant variables
+    result_vars = OrderedDict()
+    result_coord_names = set(datasets[0].coords)
+    result_attrs = datasets[0].attrs
+
+    for k, v in datasets[0].variables.items():
+        if k not in concat_over:
+            insert_result_variable(k, v)
+
+    # we've already verified everything is consistent; now, calculate
+    # shared dimension sizes so we can expand the necessary variables
+    dim_lengths = [ds.dims.get(dim, 1) for ds in datasets]
+    non_concat_dims = {}
+    for ds in datasets:
+        non_concat_dims.update(ds.dims)
+    non_concat_dims.pop(dim, None)
+
+    def ensure_common_dims(vars):
+        # ensure each variable with the given name shares the same
+        # dimensions and the same shape for all of them except along the
+        # concat dimension
+        common_dims = tuple(pd.unique([d for v in vars for d in v.dims]))
+        if dim not in common_dims:
+            common_dims = (dim,) + common_dims
+        for var, dim_len in zip(vars, dim_lengths):
+            if var.dims != common_dims:
+                common_shape = tuple(non_concat_dims.get(d, dim_len)
+                                     for d in common_dims)
+                var = var.expand_dims(common_dims, common_shape)
+            yield var
+
+    # stack up each variable to fill-out the dataset
+    for k in concat_over:
+        vars = ensure_common_dims([ds.variables[k] for ds in datasets])
+        combined = concat_vars(vars, dim, positions)
+        insert_result_variable(k, combined)
+
+    result = Dataset(result_vars, attrs=result_attrs)
+    result = result.set_coords(result_coord_names)
+
+    if coord is not None:
+        # add concat dimension last to ensure that its in the final Dataset
+        result[coord.name] = coord
+
+    return result
 
 
 def _generate_fake_points(components, statevar_dict, energy_limit, output, maximum_internal_dof, broadcast):
@@ -61,8 +149,83 @@ def _generate_fake_points(components, statevar_dict, energy_limit, output, maxim
     return Dataset(data_arrays, coords=coordinate_dict)
 
 
+@cacheit
+def _sample_phase_constitution(phase_name, phase_constituents, sublattice_dof, comps,
+                               variables, sampler, fixed_grid, pdens):
+    """
+    Sample the internal degrees of freedom of a phase.
+
+    Parameters
+    ----------
+    phase_name
+    phase_constituents
+    sublattice_dof
+    comps
+    variables
+    sampler
+    fixed_grid
+    pdens
+
+    Returns
+    -------
+    ndarray of points
+    """
+    # Eliminate pure vacancy endmembers from the calculation
+    vacancy_indices = list()
+    for idx, sublattice in enumerate(phase_constituents):
+        active_in_subl = sorted(set(phase_constituents[idx]).intersection(comps))
+        if 'VA' in active_in_subl and 'VA' in sorted(comps):
+            vacancy_indices.append(active_in_subl.index('VA'))
+    if len(vacancy_indices) != len(phase_constituents):
+        vacancy_indices = None
+    # Add all endmembers to guarantee their presence
+    points = endmember_matrix(sublattice_dof,
+                              vacancy_indices=vacancy_indices)
+    if fixed_grid is True:
+        # Sample along the edges of the endmembers
+        # These constitution space edges are often the equilibrium points!
+        em_pairs = list(itertools.combinations(points, 2))
+        lingrid = np.linspace(0, 1, pdens)
+        extra_points = [first_em * lingrid[np.newaxis].T +
+                        second_em * lingrid[::-1][np.newaxis].T
+                        for first_em, second_em in em_pairs]
+        points = np.concatenate(list(itertools.chain([points], extra_points)))
+
+    # Sample composition space for more points
+    if sum(sublattice_dof) > len(sublattice_dof):
+        points = np.concatenate((points,
+                                 sampler(sublattice_dof,
+                                         pdof=pdens)
+                                 ))
+
+    # If there are nontrivial sublattices with vacancies in them,
+    # generate a set of points where their fraction is zero and renormalize
+    for idx, sublattice in enumerate(phase_constituents):
+        if 'VA' in set(sublattice) and len(sublattice) > 1:
+            var_idx = variables.index(v.SiteFraction(phase_name, idx, 'VA'))
+            addtl_pts = np.copy(points)
+            # set vacancy fraction to log-spaced between 1e-10 and 1e-6
+            addtl_pts[:, var_idx] = np.power(10.0, -10.0 * (1.0 - addtl_pts[:, var_idx]))
+            # renormalize site fractions
+            cur_idx = 0
+            for ctx in sublattice_dof:
+                end_idx = cur_idx + ctx
+                addtl_pts[:, cur_idx:end_idx] /= \
+                    addtl_pts[:, cur_idx:end_idx].sum(axis=1)[:, None]
+                cur_idx = end_idx
+            # add to points matrix
+            points = np.concatenate((points, addtl_pts), axis=0)
+    # Filter out nan's that may have slipped in if we sampled too high a vacancy concentration
+    # Issues with this appear to be platform-dependent
+    points = points[~np.isnan(points).any(axis=-1)]
+    # Ensure that points has the correct dimensions and dtype
+    points = np.atleast_2d(np.asarray(points, dtype=np.float))
+    return points
+
+
 def _compute_phase_values(phase_obj, components, variables, statevar_dict,
-                          points, func, output, maximum_internal_dof, broadcast=True):
+                          points, phase_record, output, maximum_internal_dof, broadcast=True, fake_points=False,
+                          largest_energy=None):
     """
     Calculate output values for a particular phase.
 
@@ -78,9 +241,8 @@ def _compute_phase_values(phase_obj, components, variables, statevar_dict,
         Mapping of state variables to desired values. This will broadcast if necessary.
     points : ndarray
         Inputs to 'func', except state variables. Columns should be in 'variables' order.
-    func : callable
-        Function of state variables and 'variables'.
-        See 'make_callable' docstring for details.
+    phase_record : PhaseRecord
+        Contains callable for energy and phase metadata.
     output : string
         Desired name of the output result in the Dataset.
     maximum_internal_dof : int
@@ -89,6 +251,10 @@ def _compute_phase_values(phase_obj, components, variables, statevar_dict,
     broadcast : bool
         If True, broadcast state variables against each other to create a grid.
         If False, assume state variables are given as equal-length lists (or single-valued).
+    fake_points : bool, optional (Default: False)
+        If True, the first few points of the output surface will be fictitious
+        points used to define an equilibrium hyperplane guaranteed to be above
+        all the other points. This is used for convex hull computations.
 
     Returns
     -------
@@ -116,10 +282,23 @@ def _compute_phase_values(phase_obj, components, variables, statevar_dict,
                                  'broadcast=False.')
             statevars_.append(statevar)
         statevars = statevars_
-    phase_output = func(*itertools.chain(statevars, np.rollaxis(points, -1, start=0)))
+    # func may only have support for vectorization along a single axis (no broadcasting)
+    # we need to force broadcasting and flatten the result before calling
+    bc_statevars = [np.ascontiguousarray(broadcast_to(x, points.shape[:-1]).reshape(-1)) for x in statevars]
+    pts = points.reshape(-1, points.shape[-1]).T
+    dof = np.ascontiguousarray(np.concatenate((bc_statevars, pts), axis=0).T)
+    phase_output = np.ascontiguousarray(np.zeros(dof.shape[0]))
+    phase_record.obj(phase_output, dof)
     if isinstance(phase_output, (float, int)):
         phase_output = broadcast_to(phase_output, points.shape[:-1])
     phase_output = np.asarray(phase_output, dtype=np.float)
+    phase_output.shape = points.shape[:-1]
+    if fake_points:
+        phase_output = np.concatenate((broadcast_to(largest_energy, points.shape[:-2] + (len(components),)), phase_output), axis=-1)
+        phase_names = np.concatenate((broadcast_to('_FAKE_', points.shape[:-2] + (len(components),)),
+                                      np.full(points.shape[:-1], phase_obj.name, dtype='U' + str(len(phase_obj.name)))), axis=-1)
+    else:
+        phase_names = np.full(points.shape[:-1], phase_obj.name, dtype='U'+str(len(phase_obj.name)))
 
     # Map the internal degrees of freedom to global coordinates
     # Normalize site ratios by the sum of site ratios times a factor
@@ -138,22 +317,27 @@ def _compute_phase_values(phase_obj, components, variables, statevar_dict,
             phase_obj.sublattices[vxx.sublattice_index] for vxx in variables]
         phase_compositions[..., :, col] = np.divide(np.dot(points[..., :, :], avector),
                                                site_ratio_normalization)
+    if fake_points:
+        phase_compositions = np.concatenate((np.broadcast_to(np.eye(len(components)), points.shape[:-2] + (len(components), len(components))), phase_compositions), axis=-2)
 
     coordinate_dict = {'component': components}
     # Resize 'points' so it has the same number of columns as the maximum
     # number of internal degrees of freedom of any phase in the calculation.
     # We do this so that everything is aligned for concat.
     # Waste of memory? Yes, but the alternatives are unclear.
-    expanded_points = np.full(points.shape[:-1] + (maximum_internal_dof,), np.nan)
-    expanded_points[..., :points.shape[-1]] = points
+    if fake_points:
+        expanded_points = np.full(points.shape[:-2] + (len(components)+points.shape[-2], maximum_internal_dof), np.nan)
+        expanded_points[..., len(components):, :points.shape[-1]] = points
+    else:
+        expanded_points = np.full(points.shape[:-1] + (maximum_internal_dof,), np.nan)
+        expanded_points[..., :points.shape[-1]] = points
     if broadcast:
         coordinate_dict.update({key: np.atleast_1d(value) for key, value in statevar_dict.items()})
         output_columns = [str(x) for x in statevar_dict.keys()] + ['points']
     else:
         output_columns = ['points']
     data_arrays = {'X': (output_columns + ['component'], phase_compositions),
-                   'Phase': (output_columns,
-                             np.full(points.shape[:-1], phase_obj.name, dtype='U'+str(len(phase_obj.name)))),
+                   'Phase': (output_columns, phase_names),
                    'Y': (output_columns + ['internal_dof'], expanded_points),
                    output: (['dim_'+str(i) for i in range(len(phase_output.shape) - len(output_columns))] + output_columns, phase_output)
                    }
@@ -164,8 +348,8 @@ def _compute_phase_values(phase_obj, components, variables, statevar_dict,
 
     return Dataset(data_arrays, coords=coordinate_dict)
 
-@TempfileManager(os.getcwd())
-def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, broadcast=True, tmpman=None, **kwargs):
+
+def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, broadcast=True, parameters=None, **kwargs):
     """
     Sample the property surface of 'output' containing the specified
     components and phases. Model parameters are taken from 'dbf' and any
@@ -190,13 +374,12 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
     broadcast : bool, optional
         If True, broadcast given state variable lists against each other to create a grid.
         If False, assume state variables are given as equal-length lists.
-    tmpman : TempfileManager, optional
-        Context manager for temporary file creation during the calculation.
     points : ndarray or a dict of phase names to ndarray, optional
         Columns of ndarrays must be internal degrees of freedom (site fractions), sorted.
         If this is not specified, points will be generated automatically.
     pdens : int, a dict of phase names to int, or a seq of both, optional
         Number of points to sample per degree of freedom.
+        Default: 2000; Default when called from equilibrium(): 500
     model : Model, a dict of phase names to Model, or a seq of both, optional
         Model class to use for each phase.
     sampler : callable, a dict of phase names to callable, or a seq of both, optional
@@ -205,6 +388,8 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
     grid_points : bool, a dict of phase names to bool, or a seq of both, optional (Default: True)
         Whether to add evenly spaced points between end-members.
         The density of points is determined by 'pdens'
+    parameters : dict, optional
+        Maps SymPy Symbol to numbers, for overriding the values of parameters in the Database.
 
     Returns
     -------
@@ -218,10 +403,15 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
     # there may be keyword arguments that aren't state variables
     pdens_dict = unpack_kwarg(kwargs.pop('pdens', 2000), default_arg=2000)
     points_dict = unpack_kwarg(kwargs.pop('points', None), default_arg=None)
-    model_dict = unpack_kwarg(kwargs.pop('model', Model), default_arg=Model)
+    model_dict = unpack_kwarg(kwargs.pop('model', FallbackModel), default_arg=FallbackModel)
     callable_dict = unpack_kwarg(kwargs.pop('callables', None), default_arg=None)
     sampler_dict = unpack_kwarg(kwargs.pop('sampler', None), default_arg=None)
     fixedgrid_dict = unpack_kwarg(kwargs.pop('grid_points', True), default_arg=True)
+    parameters = parameters or dict()
+    if isinstance(parameters, dict):
+        parameters = OrderedDict(sorted(parameters.items(), key=str))
+    param_symbols = tuple(parameters.keys())
+    param_values = np.atleast_1d(np.array(list(parameters.values()), dtype=np.float))
     if isinstance(phases, str):
         phases = [phases]
     if isinstance(comps, str):
@@ -234,11 +424,16 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
     # If we don't do this, sympy will get confused during substitution
     statevar_dict = collections.OrderedDict((v.StateVariable(key), unpack_condition(value)) \
                                             for (key, value) in sorted(kwargs.items()))
+    # XXX: CompiledModel assumes P, T are the only state variables
+    if statevar_dict.get(v.P, None) is None:
+        statevar_dict[v.P] = 101325
+    if statevar_dict.get(v.T, None) is None:
+        statevar_dict[v.T] = 300
     str_statevar_dict = collections.OrderedDict((str(key), unpack_condition(value)) \
                                                 for (key, value) in statevar_dict.items())
     all_phase_data = []
     comp_sets = {}
-    largest_energy = -np.inf
+    largest_energy = 1e30
     maximum_internal_dof = 0
 
     # Consider only the active phases
@@ -261,12 +456,7 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
                                phase_name)
                 continue
         if points_dict[phase_name] is None:
-            try:
-                out = getattr(mod, output)
-                maximum_internal_dof = max(maximum_internal_dof, len(out.atoms(v.SiteFraction)))
-            except AttributeError:
-                raise AttributeError('Missing Model attribute {0} specified for {1}'
-                                     .format(output, mod.__class__))
+            maximum_internal_dof = max(maximum_internal_dof, sum(len(x) for x in mod.constituents))
         else:
             maximum_internal_dof = max(maximum_internal_dof, np.asarray(points_dict[phase_name]).shape[-1])
 
@@ -275,107 +465,60 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
             mod = model_dict[phase_name]
         except KeyError:
             continue
-        # Construct an ordered list of the variables
-        variables, sublattice_dof = generate_dof(phase_obj, mod.components)
-
-        # Build the "fast" representation of that model
-        if callable_dict[phase_name] is None:
-            out = getattr(mod, output)
-            # As a last resort, treat undefined symbols as zero
-            # But warn the user when we do this
-            # This is consistent with TC's behavior
-            undefs = list(out.atoms(Symbol) - out.atoms(v.StateVariable))
-            for undef in undefs:
-                out = out.xreplace({undef: float(0)})
-                logger.warning('Setting undefined symbol %s for phase %s to zero',
-                               undef, phase_name)
-            comp_sets[phase_name] = build_functions(out, list(statevar_dict.keys()) + variables, tmpman=tmpman,
-                                                    include_obj=True, include_grad=False, include_hess=False)
+        # this is a phase model we couldn't construct for whatever reason; skip it
+        if isinstance(mod, type):
+            continue
+        if (not isinstance(mod, CompiledModel)) or (output != 'GM'):
+            if isinstance(mod, CompiledModel):
+                mod = Model(dbf, comps, phase_name, parameters=parameters)
+            # Construct an ordered list of the variables
+            variables, sublattice_dof = generate_dof(phase_obj, mod.components)
+            # Build the "fast" representation of that model
+            if callable_dict[phase_name] is None:
+                try:
+                    out = getattr(mod, output)
+                except AttributeError:
+                    raise AttributeError('Missing Model attribute {0} specified for {1}'
+                                         .format(output, mod.__class__))
+                # As a last resort, treat undefined symbols as zero
+                # But warn the user when we do this
+                # This is consistent with TC's behavior
+                undefs = list(out.atoms(Symbol) - out.atoms(v.StateVariable))
+                for undef in undefs:
+                    out = out.xreplace({undef: float(0)})
+                    logger.warning('Setting undefined symbol %s for phase %s to zero',
+                                   undef, phase_name)
+                comp_sets[phase_name] = build_functions(out, list(statevar_dict.keys()) + variables,
+                                                        include_obj=True, include_grad=False, include_hess=False,
+                                                        parameters=param_symbols)
+            else:
+                comp_sets[phase_name] = callable_dict[phase_name]
+            phase_record = PhaseRecord_from_cython(comps, list(statevar_dict.keys()) + variables,
+                                        np.array(dbf.phases[phase_name].sublattices, dtype=np.float),
+                                        param_values, comp_sets[phase_name], None, None)
         else:
-            comp_sets[phase_name] = callable_dict[phase_name]
-
+            variables = sorted(set(mod.variables) - {v.T, v.P}, key=str)
+            sublattice_dof = mod.sublattice_dof
+            phase_record = PhaseRecord_from_compiledmodel(mod, param_values)
         points = points_dict[phase_name]
         if points is None:
-            # Eliminate pure vacancy endmembers from the calculation
-            vacancy_indices = list()
-            for idx, sublattice in enumerate(phase_obj.constituents):
-                active_in_subl = sorted(set(phase_obj.constituents[idx]).intersection(comps))
-                if 'VA' in active_in_subl and 'VA' in sorted(comps):
-                    vacancy_indices.append(active_in_subl.index('VA'))
-            if len(vacancy_indices) != len(phase_obj.constituents):
-                vacancy_indices = None
-            logger.debug('vacancy_indices: %s', vacancy_indices)
-            # Add all endmembers to guarantee their presence
-            points = endmember_matrix(sublattice_dof,
-                                      vacancy_indices=vacancy_indices)
-            if fixedgrid_dict[phase_name] is True:
-                # Sample along the edges of the endmembers
-                # These constitution space edges are often the equilibrium points!
-                em_pairs = list(itertools.combinations(points, 2))
-                for first_em, second_em in em_pairs:
-                    extra_points = first_em * np.linspace(0, 1, pdens_dict[phase_name])[np.newaxis].T + \
-                                   second_em * np.linspace(0, 1, pdens_dict[phase_name])[::-1][np.newaxis].T
-                    points = np.concatenate((points, extra_points))
-
-
-            # Sample composition space for more points
-            if sum(sublattice_dof) > len(sublattice_dof):
-                sampler = sampler_dict[phase_name]
-                if sampler is None:
-                    sampler = point_sample
-                points = np.concatenate((points,
-                                         sampler(sublattice_dof,
-                                                 pdof=pdens_dict[phase_name])
-                                         ))
-
-            # If there are nontrivial sublattices with vacancies in them,
-            # generate a set of points where their fraction is zero and renormalize
-            for idx, sublattice in enumerate(phase_obj.constituents):
-                if 'VA' in set(sublattice) and len(sublattice) > 1:
-                    var_idx = variables.index(v.SiteFraction(phase_name, idx, 'VA'))
-                    addtl_pts = np.copy(points)
-                    # set vacancy fraction to log-spaced between 1e-10 and 1e-6
-                    addtl_pts[:, var_idx] = np.power(10.0, -10.0*(1.0 - addtl_pts[:, var_idx]))
-                    # renormalize site fractions
-                    cur_idx = 0
-                    for ctx in sublattice_dof:
-                        end_idx = cur_idx + ctx
-                        addtl_pts[:, cur_idx:end_idx] /= \
-                            addtl_pts[:, cur_idx:end_idx].sum(axis=1)[:, None]
-                        cur_idx = end_idx
-                    # add to points matrix
-                    points = np.concatenate((points, addtl_pts), axis=0)
-            # Filter out nan's that may have slipped in if we sampled too high a vacancy concentration
-            # Issues with this appear to be platform-dependent
-            points = points[~np.isnan(points).any(axis=-1)]
-        # Ensure that points has the correct dimensions and dtype
-        points = np.atleast_2d(np.asarray(points, dtype=np.float))
+            points = _sample_phase_constitution(phase_name, phase_obj.constituents, sublattice_dof, comps,
+                                                tuple(variables), sampler_dict[phase_name] or point_sample,
+                                                fixedgrid_dict[phase_name], pdens_dict[phase_name])
+        points = np.atleast_2d(points)
 
+        fp = fake_points and (phase_name == sorted(active_phases.keys())[0])
         phase_ds = _compute_phase_values(phase_obj, components, variables, str_statevar_dict,
-                                         points, comp_sets[phase_name], output,
-                                         maximum_internal_dof, broadcast=broadcast)
-        # largest_energy is really only relevant if fake_points is set
-        if fake_points:
-            largest_energy = max(phase_ds[output].max(), largest_energy)
+                                         points, phase_record, output,
+                                         maximum_internal_dof, broadcast=broadcast,
+                                         largest_energy=float(largest_energy), fake_points=fp)
         all_phase_data.append(phase_ds)
 
-    if fake_points:
-        if output != 'GM':
-            raise ValueError('fake_points=True should only be used with output=\'GM\'')
-        phase_ds = _generate_fake_points(components, statevar_dict, largest_energy, output,
-                                         maximum_internal_dof, broadcast)
-        final_ds = concat(itertools.chain([phase_ds], all_phase_data),
-                          dim='points')
+    # speedup for single-phase case (found by profiling)
+    if len(all_phase_data) > 1:
+        final_ds = _fast_concat(all_phase_data, dim='points')
+        final_ds['points'].values = np.arange(len(final_ds['points']))
+        final_ds.coords['points'].values = np.arange(len(final_ds['points']))
     else:
-        # speedup for single-phase case (found by profiling)
-        if len(all_phase_data) > 1:
-            final_ds = concat(all_phase_data, dim='points')
-        else:
-            final_ds = all_phase_data[0]
-
-    if (not fake_points) and (len(all_phase_data) == 1):
-        pass
-    else:
-        # Reset the points dimension to use a single global index
-        final_ds['points'] = np.arange(len(final_ds.points))
+        final_ds = all_phase_data[0]
     return final_ds
